@@ -22,7 +22,7 @@
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/exec.c,v 1.91 2011/05/07 00:51:11 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/exec.c,v 1.94 2011/07/07 21:24:52 tg Exp $");
 
 #ifndef MKSH_DEFAULT_EXECSHELL
 #define MKSH_DEFAULT_EXECSHELL	"/bin/sh"
@@ -405,15 +405,30 @@ execute(struct op * volatile t,
 		break;
 
 	case TCASE:
+		i = 0;
 		ccp = evalstr(t->str, DOTILDE);
-		for (t = t->left; t != NULL && t->type == TPAT; t = t->right)
-		    for (ap = (const char **)t->vars; *ap; ap++)
-			if ((s = evalstr(*ap, DOTILDE|DOPAT)) &&
-			    gmatchx(ccp, s, false))
-				goto Found;
-		break;
- Found:
-		rv = execute(t->left, flags & XERROK, xerrok);
+		for (t = t->left; t != NULL && t->type == TPAT; t = t->right) {
+			for (ap = (const char **)t->vars; *ap; ap++) {
+				if (i || ((s = evalstr(*ap, DOTILDE|DOPAT)) &&
+				    gmatchx(ccp, s, false))) {
+					rv = execute(t->left, flags & XERROK,
+					    xerrok);
+					i = 0;
+					switch (t->u.charflag) {
+					case '&':
+						i = 1;
+						/* FALLTHROUGH */
+					case '|':
+						goto TCASE_next;
+					}
+					goto TCASE_out;
+				}
+			}
+			i = 0;
+ TCASE_next:
+			/* empty */;
+		}
+ TCASE_out:
 		break;
 
 	case TBRACE:
@@ -573,6 +588,7 @@ comexec(struct op *t, struct tbl * volatile tp, const char **ap,
 				subst_exstat = 0;
 				break;
 			}
+#ifndef MKSH_NO_EXTERNAL_CAT
 		} else if (tp->val.f == c_cat) {
 			/*
 			 * if we have any flags, do not use the builtin
@@ -590,6 +606,7 @@ comexec(struct op *t, struct tbl * volatile tp, const char **ap,
 			else
 				/* go on, use the builtin */
 				break;
+#endif
 		} else
 			break;
 		tp = findcom(ap[0], fcflags & (FC_BI|FC_FUNC));
@@ -654,7 +671,7 @@ comexec(struct op *t, struct tbl * volatile tp, const char **ap,
 	/* function call */
 	case CFUNC: {
 		volatile unsigned char old_xflag;
-		volatile Tflag old_inuse;
+		volatile uint32_t old_inuse;
 		const char * volatile old_kshname;
 
 		if (!(tp->flag & ISSET)) {
@@ -1018,7 +1035,7 @@ const char *
 builtin(const char *name, int (*func) (const char **))
 {
 	struct tbl *tp;
-	Tflag flag;
+	uint32_t flag;
 
 	/* see if any flags should be set for this builtin */
 	for (flag = 0; ; name++) {
