@@ -22,7 +22,7 @@
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/jobs.c,v 1.77 2011/03/13 01:20:20 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/jobs.c,v 1.78.2.2 2011/07/16 23:37:59 tg Exp $");
 
 #if HAVE_KILLPG
 #define mksh_killpg		killpg
@@ -353,6 +353,24 @@ j_change(void)
 }
 #endif
 
+#if HAVE_NICE
+/* run nice(3) and ignore the result */
+static void
+ksh_nice(int ness)
+{
+#if defined(__USE_FORTIFY_LEVEL) && (__USE_FORTIFY_LEVEL > 0)
+	int e;
+
+	errno = 0;
+	/* this is gonna annoy users; complain to your distro, people! */
+	if (nice(ness) == -1 && (e = errno) != 0)
+		warningf(false, "%s: %s", "bgnice", strerror(e));
+#else
+	(void)nice(ness);
+#endif
+}
+#endif
+
 /* execute tree in child subprocess */
 int
 exchild(struct op *t, int flags,
@@ -499,7 +517,7 @@ exchild(struct op *t, int flags,
 #endif
 #if HAVE_NICE
 		if (Flag(FBGNICE) && (flags & XBGND))
-			(void)nice(4);
+			ksh_nice(4);
 #endif
 		if ((flags & XBGND)
 #ifndef MKSH_UNEMPLOYED
@@ -861,25 +879,6 @@ j_stopped_running(void)
 	return (0);
 }
 
-int
-j_njobs(void)
-{
-	Job *j;
-	int nj = 0;
-#ifndef MKSH_NOPROSPECTOFWORK
-	sigset_t omask;
-
-	sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
-#endif
-	for (j = job_list; j; j = j->next)
-		nj++;
-
-#ifndef MKSH_NOPROSPECTOFWORK
-	sigprocmask(SIG_SETMASK, &omask, NULL);
-#endif
-	return (nj);
-}
-
 
 /* list jobs for jobs built-in */
 int
