@@ -1,5 +1,5 @@
 /*-
- * Copyright (c) 2009, 2010
+ * Copyright (c) 2009, 2010, 2011
  *	Thorsten Glaser <tg@mirbsd.org>
  *
  * Provided that these terms and disclaimer and all copyright notices
@@ -20,13 +20,13 @@
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/lalloc.c,v 1.13 2010/09/14 21:26:14 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/lalloc.c,v 1.17 2011/03/13 10:50:44 tg Exp $");
 
 /* build with CPPFLAGS+= -DUSE_REALLOC_MALLOC=0 on ancient systems */
 #if defined(USE_REALLOC_MALLOC) && (USE_REALLOC_MALLOC == 0)
-#define remalloc(p,n)	((p) == NULL ? malloc(n) : realloc((p), (n)))
+#define remalloc(p,n)	((p) == NULL ? malloc_osi(n) : realloc_osi((p), (n)))
 #else
-#define remalloc(p,n)	realloc((p), (n))
+#define remalloc(p,n)	realloc_osi((p), (n))
 #endif
 
 #define ALLOC_ISUNALIGNED(p) (((ptrdiff_t)(p)) % ALLOC_SIZE)
@@ -61,8 +61,15 @@ findptr(ALLOC_ITEM **lpp, char *ptr, Area *ap)
 #ifndef MKSH_SMALL
  fail:
 #endif
+#ifdef DEBUG
+			internal_warningf("rogue pointer %lX in ap %lX",
+			    (long)(ptrdiff_t)ptr, (long)(ptrdiff_t)ap);
+			/* try to get a coredump */
+			abort();
+#else
 			internal_errorf("rogue pointer %lX",
 			    (long)(ptrdiff_t)ptr);
+#endif
 		}
 	return (ap);
 }
@@ -70,7 +77,7 @@ findptr(ALLOC_ITEM **lpp, char *ptr, Area *ap)
 void *
 aresize2(void *ptr, size_t fac1, size_t fac2, Area *ap)
 {
-	if (fac1 && fac2 && (SIZE_MAX / fac1 < fac2))
+	if (notoktomul(fac1, fac2))
 		internal_errorf(T_intovfl, (unsigned long)fac1, '*',
 		    (unsigned long)fac2);
 	return (aresize(ptr, fac1 * fac2, ap));
@@ -95,8 +102,7 @@ aresize(void *ptr, size_t numb, Area *ap)
 	    || ALLOC_ISUNALIGNED(lp)
 #endif
 	    )
-		internal_errorf("can't allocate %lu data bytes",
-		    (unsigned long)numb);
+		internal_errorf(T_oomem, (unsigned long)numb);
 	/* this only works because Area is an ALLOC_ITEM */
 	lp->next = ap->next;
 	ap->next = lp;
@@ -114,7 +120,7 @@ afree(void *ptr, Area *ap)
 		/* unhook */
 		pp->next = lp->next;
 		/* now free ALLOC_ITEM */
-		free(lp);
+		free_osimalloc(lp);
 	}
 }
 
@@ -128,6 +134,6 @@ afreeall(Area *ap)
 		/* make next ALLOC_ITEM head of list */
 		ap->next = lp->next;
 		/* free old head */
-		free(lp);
+		free_osimalloc(lp);
 	}
 }
