@@ -1,5 +1,5 @@
 #if defined(SHFLAGS_DEFNS)
-__RCSID("$MirOS: src/bin/mksh/sh_flags.h,v 1.4 2009/10/02 18:08:37 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/sh_flags.h,v 1.8 2011/02/11 01:18:23 tg Exp $");
 #define FN(sname,cname,ochar,flags)	/* nothing */
 #elif defined(SHFLAGS_ENUMS)
 #define FN(sname,cname,ochar,flags)	cname,
@@ -21,10 +21,8 @@ __RCSID("$MirOS: src/bin/mksh/sh_flags.h,v 1.4 2009/10/02 18:08:37 tg Exp $");
 /* -a	all new parameters are created with the export attribute */
 F0("allexport", FEXPORT, 'a', OF_ANY)
 
-#if HAVE_ARC4RANDOM
-/* ./.	backwards compat: available if arc4random(3) is used for $RANDOM */
-FN("arc4random", FARC4RANDOM, 0, OF_INTERNAL)
-#endif
+/* ./.	backwards compat: dummy, emits a warning */
+FN("arc4random", FARC4RANDOM, 0, OF_ANY)
 
 #if HAVE_NICE
 /* ./.	bgnice */
@@ -89,13 +87,16 @@ FN("nounset", FNOUNSET, 'u', OF_ANY)
 /* ./.	don't do logical cds/pwds (non-standard) */
 FN("physical", FPHYSICAL, 0, OF_ANY)
 
+/* ./.	pdksh compat: somewhat more POSIXish mode (non-standard) */
+FN("posix", FPOSIX, 0, OF_ANY)
+
 /* -p	use suid_profile; privileged shell */
 FN("privileged", FPRIVILEGED, 'p', OF_ANY)
 
 /* -r	restricted shell */
 FN("restricted", FRESTRICTED, 'r', OF_CMDLINE)
 
-/* ./.	called as sh (some POSIX, some kludgy) not mksh (non-standard) */
+/* ./.	pdksh compat: called as sh not mksh; kludge mode (non-standard) */
 FN("sh", FSH, 0, OF_ANY)
 
 /* -s	(invocation) parse stdin (pseudo non-standard) */
@@ -134,6 +135,9 @@ FN(NULL, FCOMMAND, 'c', OF_CMDLINE)
  * anonymous flags: used internally by shell only (not visible to user)
  */
 
+/* ./.	direct builtin call (divined from argv[0] multi-call binary) */
+FN(NULL, FAS_BUILTIN, 0, OF_INTERNAL)
+
 /* ./.	(internal) initial shell was interactive */
 FN(NULL, FTALKING_I, 0, OF_INTERNAL)
 
